#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

# Python imports
from enum import Enum
import logging

# C imports
from libc cimport stdint

cimport c_message
cimport c_amqp_definitions
cimport c_amqpvalue


_logger = logging.getLogger(__name__)


class MessageBodyType(Enum):
    NoneType = c_message.MESSAGE_BODY_TYPE_TAG.MESSAGE_BODY_TYPE_NONE
    DataType = c_message.MESSAGE_BODY_TYPE_TAG.MESSAGE_BODY_TYPE_DATA
    SequenceType = c_message.MESSAGE_BODY_TYPE_TAG.MESSAGE_BODY_TYPE_SEQUENCE
    ValueType = c_message.MESSAGE_BODY_TYPE_TAG.MESSAGE_BODY_TYPE_VALUE


cdef message_factory(c_message.MESSAGE_HANDLE value):
    new_message = cMessage()
    new_message.wrap(value)
    return new_message


cpdef create_message():
    new_message = cMessage()
    new_message.create()
    return new_message


cdef class cMessage(StructBase):

    cdef c_message.MESSAGE_HANDLE _c_value

    def __cinit__(self):
        pass

    def __dealloc__(self):
        _logger.debug("Deallocating {}".format(self.__class__.__name__))
        self.destroy()

    cdef _create(self):
        if <void*>self._c_value is NULL:
            self._memory_error()

    cpdef destroy(self):
        if <void*>self._c_value is not NULL:
            _logger.debug("Destorying {}".format(self.__class__.__name__))
            c_message.message_destroy(self._c_value)
            self._c_value = <c_message.MESSAGE_HANDLE>NULL

    cdef wrap(self, c_message.MESSAGE_HANDLE value):
        self.destroy()
        self._c_value = value
        self._create()

    cdef create(self):
        self.destroy()
        self._c_value = c_message.message_create()
        self._create()

    cpdef clone(self):
        cdef c_message.MESSAGE_HANDLE value
        value = c_message.message_clone(self._c_value)
        if <void*>value == NULL:
            self._value_error()
        return message_factory(value)

    @property
    def header(self):
        cdef c_amqp_definitions.HEADER_HANDLE value
        if c_message.message_get_header(self._c_value, &value) == 0:
            if <void*>value == NULL:
                return None
            new_obj = cHeader()
            new_obj.wrap(value)
            return new_obj
        else:
            self._value_error()

    @header.setter
    def header(self, cHeader value):
        if value is None:
            if c_message.message_set_header(
                self._c_value, <c_amqp_definitions.HEADER_HANDLE>NULL) != 0:
                    self._value_error()
        elif c_message.message_set_header(
            self._c_value, <c_amqp_definitions.HEADER_HANDLE>value._c_value) != 0:
                self._value_error()

    @property
    def delivery_annotations(self):
        cdef c_amqp_definitions.delivery_annotations value
        if c_message.message_get_delivery_annotations(self._c_value, &value) == 0:
            if <void*>value == NULL:
                return None
            new_obj = cDeliveryAnnotations()
            new_obj.wrap(value)
            return new_obj
        else:
            self._value_error()

    @delivery_annotations.setter
    def delivery_annotations(self, cDeliveryAnnotations value):
        if value is None:
            if c_message.message_set_delivery_annotations(
                self._c_value, <c_amqp_definitions.delivery_annotations>NULL) != 0:
                    self._value_error()
        elif c_message.message_set_delivery_annotations(
            self._c_value, <c_amqp_definitions.delivery_annotations>value._c_value) != 0:
                self._value_error()

    @property
    def message_annotations(self):
        cdef c_amqp_definitions.message_annotations value
        if c_message.message_get_message_annotations(self._c_value, &value) == 0:
            if <void*>value == NULL:
                return None
            new_obj = cMessageAnnotations()
            new_obj.wrap(value)
            return new_obj
        else:
            self._value_error()

    @message_annotations.setter
    def message_annotations(self, cMessageAnnotations value):
        if value is None:
            if c_message.message_set_message_annotations(
                self._c_value, <c_amqp_definitions.message_annotations>NULL) != 0:
                    self._value_error()
        elif c_message.message_set_message_annotations(
            self._c_value, <c_amqp_definitions.message_annotations>value._c_value) != 0:
                self._value_error()

    @property
    def properties(self):
        cdef c_amqp_definitions.PROPERTIES_HANDLE value
        if c_message.message_get_properties(self._c_value, &value) == 0:
            if <void*>value == NULL:
                return None
            new_obj = cProperties()
            new_obj.wrap(value)
            return new_obj
        else:
            self._value_error()

    @properties.setter
    def properties(self, cProperties value):
        if value is None:
            if c_message.message_set_properties(
                self._c_value, <c_amqp_definitions.PROPERTIES_HANDLE>NULL) != 0:
                    self._value_error()
        elif c_message.message_set_properties(
            self._c_value, <c_amqp_definitions.PROPERTIES_HANDLE>value._c_value) != 0:
                self._value_error()

    @property
    def application_properties(self):
        cdef c_amqpvalue.AMQP_VALUE value
        if c_message.message_get_application_properties(self._c_value, &value) == 0:
            if <void*>value == NULL:
                return None
            new_obj = cApplicationProperties()
            new_obj.wrap(value)
            return new_obj
        else:
            self._value_error()

    @application_properties.setter
    def application_properties(self, AMQPValue value):
        if value is None:
            if c_message.message_set_application_properties(
                self._c_value, <c_amqpvalue.AMQP_VALUE>NULL) != 0:
                    self._value_error()
        elif c_message.message_set_application_properties(
            self._c_value, <c_amqpvalue.AMQP_VALUE>value._c_value) != 0:
                self._value_error()

    @property
    def footer(self):
        cdef c_amqp_definitions.annotations value
        if c_message.message_get_footer(self._c_value, &value) == 0:
            if <void*>value == NULL:
                return None
            new_obj = cFooter()
            new_obj.wrap(value)
            return new_obj
        else:
            self._value_error()

    @footer.setter
    def footer(self, cFooter value):
        if value is None:
            if c_message.message_set_footer(
                self._c_value, <c_amqp_definitions.annotations>NULL) != 0:
                    self._value_error()
        elif c_message.message_set_footer(
            self._c_value, <c_amqp_definitions.annotations>value._c_value) != 0:
                self._value_error()

    @property
    def body_type(self):
        cdef c_message.MESSAGE_BODY_TYPE_TAG value
        if c_message.message_get_body_type(self._c_value, &value) == 0:
            return MessageBodyType(value)
        else:
            self._value_error()

    @property
    def message_format(self):
        cdef stdint.uint32_t value
        if c_message.message_get_message_format(self._c_value, &value) == 0:
            return value
        else:
            self._value_error()

    @message_format.setter
    def message_format(self, stdint.uint32_t value):
        if c_message.message_set_message_format(self._c_value, value) != 0:
                self._value_error()

    cpdef add_body_data(self, bytes value):
        cdef c_message.BINARY_DATA _binary
        length = len(value)
        bytes = value[:length]
        _binary.length = length
        _binary.bytes = bytes
        if c_message.message_add_body_amqp_data(self._c_value, _binary) != 0:
            self._value_error()

    cpdef get_body_data(self, size_t index):
        cdef c_message.BINARY_DATA _value
        if c_message.message_get_body_amqp_data_in_place(self._c_value, index, &_value) == 0:
            return _value.bytes[:_value.length]
        else:
            self._value_error()

    cpdef count_body_data(self):
        cdef size_t body_count
        if c_message.message_get_body_amqp_data_count(self._c_value, &body_count) == 0:
            return body_count
        else:
            self._value_error()

    cpdef set_body_value(self, AMQPValue value):
        if c_message.message_set_body_amqp_value(self._c_value, <c_amqpvalue.AMQP_VALUE>value._c_value) != 0:
            self._value_error()

    cpdef get_body_value(self):
        cdef c_amqpvalue.AMQP_VALUE _value
        if c_message.message_get_body_amqp_value_in_place(self._c_value, &_value) == 0:
            return value_factory(_value)
        else:
            self._value_error()

    cpdef add_body_sequence(self, AMQPValue sequence):
        if c_message.message_add_body_amqp_sequence(self._c_value, <c_amqpvalue.AMQP_VALUE>sequence._c_value) != 0:
            self._value_error()

    cpdef get_body_sequence(self, size_t index):
        cdef c_amqpvalue.AMQP_VALUE _value
        if c_message.message_get_body_amqp_sequence_in_place(self._c_value, index, &_value) == 0:
            return value_factory(_value)
        else:
            self._value_error()

    cpdef count_body_sequence(self):
        cdef size_t body_count
        if c_message.message_get_body_amqp_sequence_count(self._c_value, &body_count) == 0:
            return body_count
        else:
            self._value_error()


cdef class Messaging:

    @staticmethod
    def create_source(const char* address):
        cdef c_amqpvalue.AMQP_VALUE _value
        _value = c_message.messaging_create_source(address)
        if <void*>_value == NULL:
            raise MemoryError("Failed to allocate memory for messaging source.")
        return value_factory(_value)

    @staticmethod
    def create_target(const char* address):
        cdef c_amqpvalue.AMQP_VALUE _value
        _value = c_message.messaging_create_target(address)
        if <void*>_value == NULL:
            raise MemoryError("Failed to allocate memory for messaging target.")
        return value_factory(_value)

    @staticmethod
    def delivery_received(stdint.uint32_t section_number, stdint.uint64_t section_offset):
        cdef c_amqpvalue.AMQP_VALUE _value
        _value = c_message.messaging_delivery_received(section_number, section_offset)
        if <void*>_value == NULL:
            raise MemoryError("Failed to allocate memory for received delivery.")
        return value_factory(_value)

    @staticmethod
    def delivery_accepted():
        cdef c_amqpvalue.AMQP_VALUE _value
        _value = c_message.messaging_delivery_accepted()
        if <void*>_value == NULL:
            raise MemoryError("Failed to allocate memory for accepted delivery.")
        return value_factory(_value)

    @staticmethod
    def delivery_rejected(const char* error_condition, const char* error_description):
        cdef c_amqpvalue.AMQP_VALUE _value
        _value = c_message.messaging_delivery_rejected(error_condition, error_description)
        if <void*>_value == NULL:
            raise MemoryError("Failed to allocate memory for rejected delivery.")
        return value_factory(_value)

    @staticmethod
    def delivery_released():
        cdef c_amqpvalue.AMQP_VALUE _value
        _value = c_message.messaging_delivery_released()
        if <void*>_value == NULL:
            raise MemoryError("Failed to allocate memory for released delivery.")
        return value_factory(_value)

    @staticmethod
    def delivery_modified(bint delivery_failed, bint undeliverable_here, cFields message_annotations):
        _logger.debug("delivery modified: {} {}".format(delivery_failed, undeliverable_here))
        cdef c_amqpvalue.AMQP_VALUE _value
        _value = c_message.messaging_delivery_modified(delivery_failed, undeliverable_here, <c_amqp_definitions.fields>message_annotations._c_value)
        if <void*>_value == NULL:
            raise MemoryError("Failed to allocate memory for modified delivery.")
        return value_factory(_value)


cpdef size_t get_encoded_message_size(cMessage message):

        cdef c_message.MESSAGE_HANDLE c_msg
        c_msg = c_message.message_clone(<c_message.MESSAGE_HANDLE>message._c_value)
        if <void*>c_msg == NULL:
            raise MemoryError("Unable to clone cMessage.")

        cdef c_amqp_definitions.HEADER_HANDLE header
        cdef c_amqpvalue.AMQP_VALUE header_amqp_value
        cdef c_amqp_definitions.PROPERTIES_HANDLE properties
        cdef c_amqpvalue.AMQP_VALUE properties_amqp_value
        cdef c_amqpvalue.AMQP_VALUE application_properties
        cdef c_amqpvalue.AMQP_VALUE application_properties_value
        cdef c_amqpvalue.AMQP_VALUE body_amqp_value
        cdef c_amqpvalue.AMQP_VALUE msg_annotations
        cdef size_t encoded_size
        cdef size_t total_encoded_size = 0
        cdef size_t body_data_count = 0

        # message header
        if c_message.message_get_header(c_msg, &header) == 0 and <void*>header != NULL:
            header_amqp_value = c_amqp_definitions.amqpvalue_create_header(header)
            if <void*>header_amqp_value == NULL:
                _logger.debug("Cannot create header AMQP value")
                raise MemoryError("Cannot get cMessage header.")
            else:
                if c_amqpvalue.amqpvalue_get_encoded_size(header_amqp_value, &encoded_size) != 0:
                    _logger.debug("Cannot obtain header encoded size")
                    raise ValueError("Cannot obtain header encoded size")
                else:
                    total_encoded_size += encoded_size

        # message annotations
        if c_message.message_get_message_annotations(c_msg, &msg_annotations) == 0 and <void*>msg_annotations != NULL:
            if c_amqpvalue.amqpvalue_get_encoded_size(msg_annotations, &encoded_size) != 0:
                _logger.debug("Cannot obtain message annotations encoded size")
                raise ValueError("Cannot obtain message annotations encoded size")
            else:
                total_encoded_size += encoded_size

        # properties
        if c_message.message_get_properties(c_msg, &properties) == 0 and <void*>properties != NULL:
            properties_amqp_value = c_amqp_definitions.amqpvalue_create_properties(properties)
            if <void*>properties_amqp_value == NULL:
                _logger.debug("Cannot create properties AMQP value")
                raise MemoryError("Cannot get cMessage properties.")
            else:
                if c_amqpvalue.amqpvalue_get_encoded_size(properties_amqp_value, &encoded_size) != 0:
                    _logger.debug("Cannot obtain message properties encoded size")
                    raise ValueError("Cannot obtain message properties encoded size")
                else:
                    total_encoded_size += encoded_size

        # application properties
        if c_message.message_get_application_properties(c_msg, &application_properties) == 0 and <void*>application_properties != NULL:
            application_properties_value = c_amqp_definitions.amqpvalue_create_application_properties(application_properties)
            if <void*>application_properties_value == NULL:
                _logger.debug("Cannot create application properties AMQP value")
                raise MemoryError("Cannot get cMessage application properties.")
            else:
                if c_amqpvalue.amqpvalue_get_encoded_size(application_properties_value, &encoded_size) != 0:
                    _logger.debug("Cannot obtain application properties encoded size")
                    raise ValueError("Cannot obtain application properties encoded size")
                else:
                    total_encoded_size += encoded_size

        return total_encoded_size


#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

# Python imports
import logging

# C imports
from libc cimport stdint
cimport c_session
cimport c_connection


_logger = logging.getLogger(__name__)


cpdef create_session(Connection connection):
    session = cSession()
    session.create(connection._c_value, NULL, NULL)
    return session


cdef class cSession(StructBase):

    _links = []
    cdef c_session.SESSION_HANDLE _c_value

    def __cinit__(self):
        pass

    def __dealloc__(self):
        _logger.debug("Deallocating {}".format(self.__class__.__name__))
        self.destroy()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.destroy()

    @property
    def incoming_window(self):
        cdef stdint.uint32_t value
        if c_session.session_get_incoming_window(self._c_value, &value) != 0:
            self._value_error()
        return value

    @incoming_window.setter
    def incoming_window(self, stdint.uint32_t value):
        if c_session.session_set_incoming_window(self._c_value, value) != 0:
            self._value_error()

    @property
    def outgoing_window(self):
        cdef stdint.uint32_t value
        if c_session.session_get_outgoing_window(self._c_value, &value) != 0:
            self._value_error()
        return value

    @outgoing_window.setter
    def outgoing_window(self, stdint.uint32_t value):
        if c_session.session_set_outgoing_window(self._c_value, value) != 0:
            self._value_error()

    @property
    def handle_max(self):
        cdef c_amqp_definitions.handle value
        if c_session.session_get_handle_max(self._c_value, &value) != 0:
            self._value_error()
        return value

    @handle_max.setter
    def handle_max(self, c_amqp_definitions.handle value):
        if c_session.session_set_handle_max(self._c_value, value) != 0:
            self._value_error()

    cdef _create(self):
        if <void*>self._c_value is NULL:
            self._memory_error()

    cpdef destroy(self):
        if <void*>self._c_value is not NULL:
            _logger.debug("Destroying {}".format(self.__class__.__name__))
            c_session.session_destroy(self._c_value)
            self._c_value = <c_session.SESSION_HANDLE>NULL

    cdef wrap(self, c_session.SESSION_HANDLE value):
        self.destroy()
        self._c_value = value
        self._create()

    cdef create(self, c_connection.CONNECTION_HANDLE connection, c_session.ON_LINK_ATTACHED on_link_attached, void* callback_context):
        self.destroy()
        self._c_value = c_session.session_create(connection, on_link_attached, callback_context)
        self._create()

    cpdef begin(self):
        if c_session.session_begin(self._c_value) != 0:
            self._value_error()

    cpdef end(self, const char* condition_value, const char* description):
        if c_session.session_end(self._c_value, condition_value, description) != 0:
            self._value_error()



#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

# pylint: disable=super-init-not-called,no-self-use

import logging
import certifi

from uamqp import constants
from uamqp import c_uamqp


_logger = logging.getLogger(__name__)


class AMQPAuth:
    """AMQP authentication mixin.

    :param hostname: The AMQP endpoint hostname.
    :type hostname: str or bytes
    :param port: The TLS port - default for AMQP is 5671.
    :type port: int
    :param verify: The path to a user-defined certificate.
    :type verify: str
    :param http_proxy: HTTP proxy configuration. This should be a dictionary with
     the following keys present: 'proxy_hostname' and 'proxy_port'. Additional optional
     keys are 'username' and 'password'.
    :type http_proxy: dict
    :param encoding: The encoding to use if hostname is provided as a str.
     Default is 'UTF-8'.
    :type encoding: str
    """

    def __init__(self, hostname, port=constants.DEFAULT_AMQPS_PORT, verify=None, http_proxy=None, encoding='UTF-8'):
        self._encoding = encoding
        self.hostname = hostname.encode(self._encoding) if isinstance(hostname, str) else hostname
        self.cert_file = verify
        self.sasl = _SASL()
        self.set_tlsio(self.hostname, port, http_proxy)

    def _build_proxy_config(self, hostname, port, proxy_settings):
        config = c_uamqp.HTTPProxyConfig()
        config.hostname = hostname
        config.port = port
        for key, value in proxy_settings.items():
            proxy_settings[key] = value.encode(self._encoding) if isinstance(value, str) else value
        config.proxy_hostname = proxy_settings['proxy_hostname']
        config.proxy_port = proxy_settings['proxy_port']
        username = proxy_settings.get('username')
        if username:
            config.username = username
        password = proxy_settings.get('password')
        if password:
            config.password = password
        return config

    def set_tlsio(self, hostname, port, http_proxy):
        """Setup the default underlying TLS IO layer. On Windows this is
        Schannel, on Linux and MacOS this is OpenSSL.

        :param hostname: The endpoint hostname.
        :type hostname: bytes
        :param port: The TLS port.
        :type port: int
        """
        _default_tlsio = c_uamqp.get_default_tlsio()
        _tlsio_config = c_uamqp.TLSIOConfig()
        _tlsio_config.hostname = hostname
        _tlsio_config.port = int(port)
        if http_proxy:
            proxy_config = self._build_proxy_config(hostname, port, http_proxy)
            _tlsio_config.set_proxy_config(proxy_config)
        self._underlying_xio = c_uamqp.xio_from_tlsioconfig(_default_tlsio, _tlsio_config)

        cert = self.cert_file or certifi.where()
        with open(cert, 'rb') as cert_handle:
            cert_data = cert_handle.read()
            try:
                self._underlying_xio.set_certificates(cert_data)
            except ValueError:
                _logger.warning('Unable to set external certificates.')
        self.sasl_client = _SASLClient(self._underlying_xio, self.sasl)

    def close(self):
        """Close the authentication layer and cleanup
        all the authentication wrapper objects.
        """
        self.sasl.mechanism.destroy()
        self.sasl_client.get_client().destroy()
        self._underlying_xio.destroy()


class SASLPlain(AMQPAuth):
    """SASL Plain AMQP authentication.
    This is SASL authentication using a basic username and password.

    :param hostname: The AMQP endpoint hostname.
    :type hostname: str or bytes
    :param username: The authentication username.
    :type username: bytes or str
    :param password: The authentication password.
    :type password: bytes or str
    :param port: The TLS port - default for AMQP is 5671.
    :type port: int
    :param verify: The path to a user-defined certificate.
    :type verify: str
    :param http_proxy: HTTP proxy configuration. This should be a dictionary with
     the following keys present: 'proxy_hostname' and 'proxy_port'. Additional optional
     keys are 'username' and 'password'.
    :type http_proxy: dict
    :param encoding: The encoding to use if hostname and credentials
     are provided as a str. Default is 'UTF-8'.
    :type encoding: str
    """

    def __init__(
            self, hostname, username, password, port=constants.DEFAULT_AMQPS_PORT,
            verify=None, http_proxy=None, encoding='UTF-8'):
        self._encoding = encoding
        self.hostname = hostname.encode(self._encoding) if isinstance(hostname, str) else hostname
        self.username = username.encode(self._encoding) if isinstance(username, str) else username
        self.password = password.encode(self._encoding) if isinstance(password, str) else password
        self.cert_file = verify
        self.sasl = _SASLPlain(self.username, self.password, encoding=self._encoding)
        self.set_tlsio(self.hostname, port, http_proxy)


class SASLAnonymous(AMQPAuth):
    """SASL Annoymous AMQP authentication mixin.
    SASL connection with no credentials. If intending to use annoymous
    auth to set up a CBS session once connected, use SASTokenAuth
    or the CBSAuthMixin instead.

    :param hostname: The AMQP endpoint hostname.
    :type hostname: str or bytes
    :param port: The TLS port - default for AMQP is 5671.
    :type port: int
    :param verify: The path to a user-defined certificate.
    :type verify: str
    :param http_proxy: HTTP proxy configuration. This should be a dictionary with
     the following keys present: 'proxy_hostname' and 'proxy_port'. Additional optional
     keys are 'username' and 'password'.
    :type http_proxy: dict
    :param encoding: The encoding to use if hostname is provided as a str.
     Default is 'UTF-8'.
    :type encoding: str
    """

    def __init__(self, hostname, port=constants.DEFAULT_AMQPS_PORT, verify=None, http_proxy=None, encoding='UTF-8'):
        self._encoding = encoding
        self.hostname = hostname.encode(self._encoding) if isinstance(hostname, str) else hostname
        self.cert_file = verify
        self.sasl = _SASLAnonymous()
        self.set_tlsio(self.hostname, port, http_proxy)


class _SASLClient:

    def __init__(self, tls_io, sasl):
        self._tls_io = tls_io
        self._sasl_mechanism = sasl.mechanism
        self._io_config = c_uamqp.SASLClientIOConfig()
        self._io_config.underlying_io = self._tls_io
        self._io_config.sasl_mechanism = self._sasl_mechanism
        self._xio = c_uamqp.xio_from_saslioconfig(self._io_config)

    def get_client(self):
        return self._xio


class _SASL:

    def __init__(self):
        self._interface = self._get_interface()
        self.mechanism = self._get_mechanism()

    def _get_interface(self):
        return None

    def _get_mechanism(self):
        return c_uamqp.get_sasl_mechanism()


class _SASLAnonymous(_SASL):

    def _get_interface(self):
        return c_uamqp.saslanonymous_get_interface()

    def _get_mechanism(self):
        return c_uamqp.get_sasl_mechanism(self._interface)


class _SASLPlain(_SASL):

    def __init__(self, authcid, passwd, authzid=None, encoding='UTF-8'):
        self._sasl_config = c_uamqp.SASLPlainConfig()
        self._sasl_config.authcid = authcid
        self._sasl_config.passwd = passwd
        if authzid:
            self._sasl_config.authzid = authzid.encode(encoding) if isinstance(authzid, str) else authzid
        super(_SASLPlain, self).__init__()

    def _get_interface(self):
        return c_uamqp.saslplain_get_interface()

    def _get_mechanism(self):
        return c_uamqp.get_plain_sasl_mechanism(self._interface, self._sasl_config)



#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

import asyncio
import logging
import functools

from uamqp import constants
from uamqp import errors
from uamqp import c_uamqp
from uamqp._async import SessionAsync
from .cbs_auth import CBSAuthMixin, SASTokenAuth


_logger = logging.getLogger(__name__)


class CBSAsyncAuthMixin(CBSAuthMixin):
    """Mixin to handle sending and refreshing CBS auth tokens asynchronously."""

    async def create_authenticator_async(self, connection, debug=False, loop=None):
        """Create the async AMQP session and the CBS channel with which
        to negotiate the token.

        :param connection: The underlying AMQP connection on which
         to create the session.
        :type connection: ~uamqp._async.connection_async.ConnectionAsync
        :param debug: Whether to emit network trace logging events for the
         CBS session. Default is `False`. Logging events are set at INFO level.
        :type debug: bool
        :param loop: A user specified event loop.
        :type loop: ~asycnio.AbstractEventLoop
        :rtype: uamqp.c_uamqp.CBSTokenAuth
        """
        self.loop = loop or asyncio.get_event_loop()
        self._async_lock = asyncio.Lock(loop=self.loop)
        self._session = SessionAsync(
            connection,
            incoming_window=constants.MAX_FRAME_SIZE_BYTES,
            outgoing_window=constants.MAX_FRAME_SIZE_BYTES,
            loop=self.loop)
        try:
            self._cbs_auth = c_uamqp.CBSTokenAuth(
                self.audience,
                self.token_type,
                self.token,
                int(self.expires_at),
                self._session._session,  # pylint: disable=protected-access
                self.timeout)
            self._cbs_auth.set_trace(debug)
        except ValueError:
            await self._session.destroy_async()
            raise errors.AMQPConnectionError(
                "Unable to open authentication session.\n"
                "Please confirm target hostname exists: {}".format(connection.hostname)) from None
        return self._cbs_auth

    async def close_authenticator_async(self):
        """Close the CBS auth channel and session asynchronously."""
        await self.loop.run_in_executor(None, functools.partial(self._cbs_auth.destroy))
        await self._session.destroy_async()

    async def handle_token_async(self):
        """This coroutine is called periodically to check the status of the current
        token if there is one, and request a new one if needed.
        If the token request fails, it will be retried according to the retry policy.
        A token refresh will be attempted if the token will expire soon.

        This function will return a tuple of two booleans. The first represents whether
        the token authentication has not completed within it's given timeout window. The
        second indicates whether the token negotiation is still in progress.

        :raises: ~uamqp.errors.AuthenticationException if the token authentication fails.
        :raises: ~uamqp.errors.TokenExpired if the token has expired and cannot be
         refreshed.
        :rtype: tuple[bool, bool]
        """
        timeout = False
        in_progress = False
        await self._async_lock.acquire()
        try:
            auth_status = await self.loop.run_in_executor(None, functools.partial(self._cbs_auth.get_status))
            auth_status = constants.CBSAuthStatus(auth_status)
            if auth_status == constants.CBSAuthStatus.Error:
                if self.retries >= self._retry_policy.retries:  # pylint: disable=no-member
                    _logger.warning("Authentication Put-Token failed. Retries exhausted.")
                    raise errors.TokenAuthFailure(*self._cbs_auth.get_failure_info())
                else:
                    error_code, error_description = self._cbs_auth.get_failure_info()
                    _logger.info("Authentication status: {}, description: {}".format(error_code, error_description))
                    _logger.info("Authentication Put-Token failed. Retrying.")
                    self.retries += 1  # pylint: disable=no-member
                    await asyncio.sleep(self._retry_policy.backoff)
                    await self.loop.run_in_executor(None, functools.partial(self._cbs_auth.authenticate))
                    in_progress = True
            elif auth_status == constants.CBSAuthStatus.Failure:
                errors.AuthenticationException("Failed to open CBS authentication link.")
            elif auth_status == constants.CBSAuthStatus.Expired:
                raise errors.TokenExpired("CBS Authentication Expired.")
            elif auth_status == constants.CBSAuthStatus.Timeout:
                timeout = True
            elif auth_status == constants.CBSAuthStatus.InProgress:
                in_progress = True
            elif auth_status == constants.CBSAuthStatus.RefreshRequired:
                _logger.info("Token will expire soon - attempting to refresh.")
                self.update_token()
                await self.loop.run_in_executor(
                    None, functools.partial(
                        self._cbs_auth.refresh,
                        self.token,
                        int(self.expires_at)))
            elif auth_status == constants.CBSAuthStatus.Idle:

                await self.loop.run_in_executor(None, functools.partial(self._cbs_auth.authenticate))
                in_progress = True
            elif auth_status != constants.CBSAuthStatus.Ok:
                raise ValueError("Invalid auth state.")
        except ValueError as e:
            raise errors.AuthenticationException(
                "Token authentication failed: {}".format(e))
        except:
            raise
        finally:
            self._async_lock.release()
        return timeout, in_progress


class SASTokenAsync(SASTokenAuth, CBSAsyncAuthMixin):
    """Asynchronous CBS authentication using SAS tokens.

    :param audience: The token audience field. For SAS tokens
     this is usually the URI.
    :type audience: str or bytes
    :param uri: The AMQP endpoint URI. This must be provided as
     a decoded string.
    :type uri: str
    :param token: The SAS token.
    :type token: str or bytes.
    :param expires_in: The total remaining seconds until the token
     expires.
    :type expires_in: ~datetime.timedelta
    :param expires_at: The timestamp at which the SAS token will expire
     formatted as seconds since epoch.
    :type expires_at: float
    :param username: The SAS token username, also referred to as the key
     name or policy name. This can optionally be encoded into the URI.
    :type username: str
    :param password: The SAS token password, also referred to as the key.
     This can optionally be encoded into the URI.
    :type password: str
    :param port: The TLS port - default for AMQP is 5671.
    :type port: int
    :param timeout: The timeout in seconds in which to negotiate the token.
     The default value is 10 seconds.
    :type timeout: int
    :param retry_policy: The retry policy for the PUT token request. The default
     retry policy has 3 retries.
    :type retry_policy: ~uamqp.authentication.cbs_auth.TokenRetryPolicy
    :param verify: The path to a user-defined certificate.
    :type verify: str
    :param token_type: The type field of the token request.
     Default value is `b"servicebus.windows.net:sastoken"`.
    :type token_type: bytes
    :param http_proxy: HTTP proxy configuration. This should be a dictionary with
     the following keys present: 'proxy_hostname' and 'proxy_port'. Additional optional
     keys are 'username' and 'password'.
    :type http_proxy: dict
    :param encoding: The encoding to use if hostname is provided as a str.
     Default is 'UTF-8'.
    :type encoding: str
    """
    pass
